/* 
 * Copyright (C) 2004-2005 Shingo W. Kagami.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: gkbinclock.c,v 1.11 2005/05/03 12:04:22 swk Exp $
 */

#ifdef GKRELLM2
#include <gkrellm2/gkrellm.h>
#else
#include <gkrellm/gkrellm.h>
#define	gkrellm_init_plugin 		  init_plugin
#define	GkrellmMonitor      		  Monitor
#define GkrellmDecal        		  Decal
#define GkrellmPanel        		  Panel
#define GkrellmStyle        		  Style
#define g_signal_connect    		  gtk_signal_connect
#define G_OBJECT            		  GTK_OBJECT
#define G_CALLBACK	    		  (GtkSignalFunc)
#define gkrellm_gtk_framed_vbox           gkrellm_framed_vbox
#define gkrellm_gtk_framed_notebook_page  gkrellm_create_framed_tab
#define gkrellm_gtk_scrolled_text_view    gkrellm_scrolled_text
#define gkrellm_gtk_text_view_append      gkrellm_add_info_text_string
#define gkrellm_gtk_check_button          gkrellm_check_button
#endif

#define	CONFIG_NAME	"gkbinclock"
#define	CONFIG_KEYWORD	"gkbinclock"
#define	STYLE_NAME	"gkbinclock"
#define MAX_LED_NO 20

#define FORMAT_NO 2
#define FORMAT_BINARY 0
#define FORMAT_BCD    1

#define LED_DECAL_NO 3

static gint style_id;
static GkrellmMonitor *monitor;
static GtkWidget *main_vbox;
static GkrellmPanel *panel;
static GkrellmDecal *led[MAX_LED_NO];
static GtkWidget *mode_24h_button;
static GtkWidget *mode_12insteadof0_button;
static GtkWidget *format_select_button[FORMAT_NO];
static GtkWidget *led_on_select_button[LED_DECAL_NO];
static GtkWidget *led_off_select_button[LED_DECAL_NO];

static gint prev_hour, prev_min, prev_sec;
static gint prev_bitpattern;
static gint mode_24h = 0;
static gint mode_12insteadof0 = 0;
static gint led_no;
static gint led_decal[3] = {D_MISC_LED0, D_MISC_LED1, D_MISC_BLANK};
static gint led_on_select  = 1;
static gint led_off_select  = 0;


#if defined(ENABLE_BCD)
static gint format_select = FORMAT_BCD;
#else
static gint format_select = FORMAT_BINARY;
#endif


static gint
panel_expose_event(GtkWidget *widget, GdkEventExpose *ev)
{
    gdk_draw_pixmap(widget->window,
		    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		    panel->pixmap, ev->area.x, ev->area.y, ev->area.x, ev->area.y,
		    ev->area.width, ev->area.height);
    return FALSE;
}

static void
cb_panel_click(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
    if (event->button == 3) {
	gkrellm_open_config_window(monitor);
    }
}

static void
cb_24h_click(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
    gtk_widget_set_sensitive(mode_12insteadof0_button,
			     !(GTK_TOGGLE_BUTTON(mode_24h_button)->active));
}

static void
set_led()
{
    struct tm *tm;
    gint i;
    gint hour, min, sec;
    gint bitpattern, bitdiff;

    tm = gkrellm_get_current_time();
    hour = tm->tm_hour;
    min = tm->tm_min;
    sec = tm->tm_sec;

    if (sec == prev_sec) {
	return; 
    }

    if (!mode_24h) {
	hour %= 12;

	/* represent 12 am/pm as 12 rather than 0 in 12h mode (if preferred) */
	if (mode_12insteadof0 && hour == 0) {
	    hour = 12;
	}
    }

    if (format_select == FORMAT_BCD) {
	bitpattern =
	      (((hour / 10) & 0x03) << 18)
	    | (((hour % 10) & 0x0f) << 14)
	    | (((min  / 10) & 0x07) << 11)
	    | (((min  % 10) & 0x0f) << 7)
	    | (((sec  / 10) & 0x07) << 4)
	    |  ((sec  % 10) & 0x0f);
    } else { /* FORMAT_BINARY */
	bitpattern
	    = ((hour & 0x1f) << 12) | ((min & 0x3f) << 6) | (sec & 0x3f);
    }


    bitdiff = bitpattern ^ prev_bitpattern;
    for (i = 0; i < led_no; i++) {
	if (bitdiff & (1 << i)) {
	    gkrellm_draw_decal_pixmap(panel, led[i],
				      (bitpattern & (1 << i))?
				      led_decal[led_on_select]:
				      led_decal[led_off_select]);
	}
    }
    prev_hour = hour;
    prev_min = min;
    prev_sec = sec;
    prev_bitpattern = bitpattern;
}

static void
update_plugin()
{
    set_led();
    gkrellm_draw_panel_layers(panel);
}

static void
locate_led_binary(gint *xs, gint *ys, gint *ptr_led_no)
{
    gint i;
    gint x, y;

    /* = indices to led[] =
     *
     *    (16)15 14 13 12    hour
     *  11 10  9  8  7  6    min
     *   5  4  3  2  1  0    sec
     *  MSB <------->  LSB
     *
     */

    *ptr_led_no = mode_24h? 17: 16;
    
    /* hour */
    x = 12; y = 3; 
    for (i = 16; i >= 12; i--) {
	xs[i] = x;
	ys[i] = y;	
	x += 8;
    }
    /* min */
    x = 4; y = 3 + 6; 
    for (i = 11; i >= 6; i--) {
	xs[i] = x;
	ys[i] = y;	
	x += 8;
    }
    /* sec */
    x = 4; y = 3 + 6 * 2; 
    for (i = 5; i >= 0; i--) {
	xs[i] = x;
	ys[i] = y;	
	x += 8;
    }
}

static void
locate_led_bcd(gint *xs, gint *ys, gint *ptr_led_no)
{
    gint i;
    gint x, y;

    /* = indices to led[] =
     *
     *     17     10     3    MSB
     *     16  13  9  6  2     :
     * (19)15  12  8  5  1     :
     *  18 14  11  7  4  0    LSB
     *  hours   mins  secs
     *
     */
    
    *ptr_led_no = mode_24h? 20: 19;

    x = 4 + 8 * 5 + 10;
    for (i = 0; i < *ptr_led_no; i++) {
	if ((i % 7) == 0) {
	    x -= 10;
	    y = 3 + 6 * 3;
	} else if ((i % 7) == 0 || (i % 7) == 4) {
	    x -= 7;
	    y = 3 + 6 * 3;
	}
	
	xs[i] = x;
	ys[i] = y;
	y -= 6;
    }
}

static void
refresh_prev_led()
{
    gint i;
    
    for (i = 0; i < led_no; i++) {
	gkrellm_draw_decal_pixmap(panel, led[i], led_decal[led_off_select]);
    }
    
    prev_hour = prev_min = prev_sec = 0;
    prev_bitpattern = 0;
}

static void
create_led_panel(gint first_create)
{
    GkrellmStyle *style;
    GdkPixmap *pixmap;
    GdkBitmap *mask;
    gint xs[MAX_LED_NO], ys[MAX_LED_NO];
    gint i;

    if (first_create) {
	if (panel != NULL) {
	    gkrellm_panel_destroy(panel);
	}
	panel = gkrellm_panel_new0();
    }

    style = gkrellm_meter_style(style_id);
    pixmap = gkrellm_decal_misc_pixmap();
    mask = gkrellm_decal_misc_mask();

    if (format_select == FORMAT_BCD) {
	locate_led_bcd(xs, ys, &led_no);
    } else { /* FORMAT_BINARY */
	locate_led_binary(xs, ys, &led_no);
    }

    for (i = 0; i < led_no; i++) {
	led[i] = gkrellm_create_decal_pixmap(panel, pixmap, mask,
					     N_MISC_DECALS, NULL,
					     xs[i], ys[i]);
    }

    refresh_prev_led();
    
    gkrellm_panel_configure(panel, NULL, style);
    gkrellm_panel_create(main_vbox, monitor, panel);

    if (first_create) {
	g_signal_connect(G_OBJECT(panel->drawing_area), "expose_event",
			   G_CALLBACK(panel_expose_event), NULL);
	g_signal_connect(G_OBJECT(panel->drawing_area), "button_press_event",
			   G_CALLBACK(cb_panel_click), NULL);
    }
}

static void
create_plugin(GtkWidget *vbox, gint first_create)
{
    main_vbox = vbox;
    create_led_panel(first_create);
}


static void
save_plugin_config(FILE *f)
{
    fprintf(f, "%s mode_24h %d\n", CONFIG_KEYWORD, mode_24h);
    fprintf(f, "%s mode_12insteadof0 %d\n", CONFIG_KEYWORD, mode_12insteadof0);
    fprintf(f, "%s format_select %d\n", CONFIG_KEYWORD, format_select);
    fprintf(f, "%s led_on_select %d\n", CONFIG_KEYWORD, led_on_select);
    fprintf(f, "%s led_off_select %d\n", CONFIG_KEYWORD, led_off_select);
}
 
static void
load_plugin_config(gchar *config_line)
{
    gchar config_keyword[32], config_data[CFG_BUFSIZE];
    gint n;

    if ((n = sscanf(config_line, "%31s %[^\n]",
		    config_keyword, config_data)) != 2) {
	return;
    }

    if (!strcmp(config_keyword, "mode_24h")) {
	sscanf(config_data, "%d", &mode_24h);
    } else if (!strcmp(config_keyword, "mode_12insteadof0")) {
	sscanf(config_data, "%d", &mode_12insteadof0);
    } else if (!strcmp(config_keyword, "format_select")) {
	sscanf(config_data, "%d", &format_select);
    } else if (!strcmp(config_keyword, "led_on_select")) {
	sscanf(config_data, "%d", &led_on_select);
    } else if (!strcmp(config_keyword, "led_off_select")) {
	sscanf(config_data, "%d", &led_off_select);
    }
}

static void
apply_plugin_config(void)
{
    gint format_new;
    gint mode_24h_new;
    gint mode_12insteadof0_new;
    gint i;

    mode_24h_new = GTK_TOGGLE_BUTTON(mode_24h_button)->active;
    mode_12insteadof0_new
	= GTK_TOGGLE_BUTTON(mode_12insteadof0_button)->active;
    
    for (i = 0; i < FORMAT_NO; i++) {
	if (GTK_TOGGLE_BUTTON(format_select_button[i])->active) {
	    format_new = i;
	}
    }

    for (i = 0; i < LED_DECAL_NO; i++) {
	if (GTK_TOGGLE_BUTTON(led_on_select_button[i])->active) {
	    led_on_select = i;
	}
	if (GTK_TOGGLE_BUTTON(led_off_select_button[i])->active) {
	    led_off_select = i;
	}
    }
    refresh_prev_led();
    
    if (mode_24h_new != mode_24h
	|| mode_12insteadof0_new != mode_12insteadof0
	|| format_new != format_select) {
	mode_24h = mode_24h_new;
	mode_12insteadof0 = mode_12insteadof0_new;
	format_select = format_new;
	create_led_panel(TRUE);
    }
}

static gchar *plugin_info_text[] =
{
    "<b>gkbinclock",
    " version 0.6\n",
    "\nCopyright (C) 2004-2005 Shingo W. Kagami. All rights reserved.\n",
    "http://www.kagami.org/gkbinclock/\n",
};


static void
create_plugin_tab(GtkWidget *tab_vbox)
{
    GtkWidget *tabs;

    GtkWidget *vbox;
    GtkWidget *vbox_format, *hbox_format;
    GtkWidget *vbox_led_on, *hbox_led_on;
    GtkWidget *vbox_led_off, *hbox_led_off;

    GtkWidget *button;
    GSList *group;
    
    GtkWidget *text;
    gint i;

    tabs = gtk_notebook_new();
    gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabs), GTK_POS_TOP);
    gtk_box_pack_start(GTK_BOX(tab_vbox), tabs, TRUE, TRUE, 0);

    /* -- Options tab -- */
    vbox = gkrellm_gtk_framed_notebook_page(tabs, "Options");

    /* 24h */
    gkrellm_gtk_check_button(vbox, &mode_24h_button, mode_24h,
			     FALSE, 0, "24h mode");
    g_signal_connect(G_OBJECT(mode_24h_button), "clicked",
		     G_CALLBACK(cb_24h_click), NULL);

    /* 12 instead of 0 */
    gkrellm_gtk_check_button(vbox, &mode_12insteadof0_button,
			     mode_12insteadof0,
			     FALSE, 0,
			     "display 12 am/pm as 12 instead of 0 (in 12h mode)");
    gtk_widget_set_sensitive(mode_12insteadof0_button, !mode_24h);
    
    /* format */
    vbox_format = gkrellm_gtk_framed_vbox(vbox, "Format",
					  4, FALSE, 0, 2);
    hbox_format = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox_format), hbox_format, FALSE, FALSE, 0);

    button = gtk_radio_button_new_with_label(NULL, "Binary");
    gtk_box_pack_start(GTK_BOX(hbox_format), button, TRUE, TRUE, 0);
    group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
    format_select_button[0] = button;
	
    button = gtk_radio_button_new_with_label(group, "BCD");
    gtk_box_pack_start(GTK_BOX(hbox_format), button, TRUE, TRUE, 0);
    group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
    format_select_button[1] = button;

    gtk_toggle_button_set_active(
	GTK_TOGGLE_BUTTON(format_select_button[format_select]), TRUE);

    /* led on */
    vbox_led_on = gkrellm_gtk_framed_vbox(vbox, "LED Decal (on)",
					  4, FALSE, 0, 2);
    hbox_led_on = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox_led_on), hbox_led_on, FALSE, FALSE, 0);

    button = gtk_radio_button_new_with_label(NULL, "D_MISC_LED0");
    gtk_box_pack_start(GTK_BOX(hbox_led_on), button, TRUE, TRUE, 0);
    group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
    led_on_select_button[0] = button;
	
    button = gtk_radio_button_new_with_label(group, "D_MISC_LED1");
    gtk_box_pack_start(GTK_BOX(hbox_led_on), button, TRUE, TRUE, 0);
    group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
    led_on_select_button[1] = button;

    button = gtk_radio_button_new_with_label(group, "D_MISC_BLANK");
    gtk_box_pack_start(GTK_BOX(hbox_led_on), button, TRUE, TRUE, 0);
    group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
    led_on_select_button[2] = button;

    gtk_toggle_button_set_active(
	GTK_TOGGLE_BUTTON(led_on_select_button[led_on_select]), TRUE);

    /* led off */
    vbox_led_off = gkrellm_gtk_framed_vbox(vbox, "LED Decal (off)",
					  4, FALSE, 0, 2);
    hbox_led_off = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox_led_off), hbox_led_off, FALSE, FALSE, 0);

    button = gtk_radio_button_new_with_label(NULL, "D_MISC_LED0");
    gtk_box_pack_start(GTK_BOX(hbox_led_off), button, TRUE, TRUE, 0);
    group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
    led_off_select_button[0] = button;
	
    button = gtk_radio_button_new_with_label(group, "D_MISC_LED1");
    gtk_box_pack_start(GTK_BOX(hbox_led_off), button, TRUE, TRUE, 0);
    group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
    led_off_select_button[1] = button;

    button = gtk_radio_button_new_with_label(group, "D_MISC_BLANK");
    gtk_box_pack_start(GTK_BOX(hbox_led_off), button, TRUE, TRUE, 0);
    group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
    led_off_select_button[2] = button;

    gtk_toggle_button_set_active(
	GTK_TOGGLE_BUTTON(led_off_select_button[led_off_select]), TRUE);


    /* -- About tab -- */
    vbox = gkrellm_gtk_framed_notebook_page(tabs, "About");
    text = gkrellm_gtk_scrolled_text_view(vbox, NULL,
					  GTK_POLICY_AUTOMATIC,
					  GTK_POLICY_AUTOMATIC);
    for (i = 0; i < sizeof(plugin_info_text) / sizeof(gchar *); i++) {
	gkrellm_gtk_text_view_append(text, plugin_info_text[i]);
    }
}


static GkrellmMonitor plugin_mon =
{
    CONFIG_NAME,   /* Name, for config tab. */
    0,		   /* Id, 0 if a plugin */
    create_plugin, /* The create function */
    update_plugin, /* The update function */
    create_plugin_tab,	   /* The config tab create function */
    apply_plugin_config,   /* Apply the config function */
    save_plugin_config,	   /* Save user config */
    load_plugin_config,	   /* Load user config */
    CONFIG_KEYWORD,	   /* config keyword */
    NULL,	   /* Undefined 2 */
    NULL,	   /* Undefined 1 */
    NULL,	   /* private */
    MON_CPU,	   /* Insert plugin before this monitor */
    NULL,   	   /* Handle if a plugin, filled in by GKrellM */
    NULL	   /* path if a plugin, filled in by GKrellM */
};


GkrellmMonitor *
gkrellm_init_plugin()
{
    style_id = gkrellm_add_meter_style(&plugin_mon, STYLE_NAME);
    monitor = &plugin_mon;
    return &plugin_mon;
}
